#!/usr/bin/env python
# -*- coding: utf-8 -*-

import serial
from serial.tools import list_ports
import asyncio


DEFAULT_PORT = 7900
MAX_BUFFER_SIZE = 42 * 10


class PeriodicTask:
    def __init__(self, func, args=None, kwargs=None):
        self.func = func
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.coro = None
        self.stop()

    def start(self, interval):
        self.interval = interval
        self.running = True
        asyncio.ensure_future(self._run())

    async def _run(self):
        while self.running:
            self.func(*self.args, **self.kwargs)
            self.coro = await asyncio.sleep(self.interval)

    def stop(self):
        self.running = False

    def __del__(self):
        self.stop()


class SerialPort:
    def __init__(self, port):
        self._buffer = b''
        self.sport = serial.Serial(port)
        self.port = port
        self.task = PeriodicTask(self.read)
        self.task.start(0.1)

    def read(self):
        self._buffer += self.sport.read(self.sport.in_waiting)

    def write(self, data):
        self.sport.write(data)

    @property
    def buffer(self):
        b = self._buffer
        self._buffer = b''
        return b


class Receiver:
    def __init__(self, sock_port, peer_host, peer_port, transport):
        self.buffer = b''
        self.sock_port = sock_port
        self.peer_port = peer_port
        self.peer_host = peer_host
        self.transport = transport


class Worker:
    def __init__(self):
        self.create_sports()
        self.receivers = set()
        self.task = PeriodicTask(self.poll)
        self.task.start(0.1)

    def poll(self):
        for tcp_port in self.sports:
            sport = self.sports[tcp_port]
            buffer = sport.buffer
            for r in self.receivers:
                if r.sock_port == tcp_port:
                    r.transport.write(buffer)

    def create_sports(self):
        self.sports = {}
        for i, sport in enumerate(list_ports.comports()):
            tcp_port = DEFAULT_PORT + i
            self.sports[tcp_port] = SerialPort(sport.device)
            print('Serial port', sport, 'on tcp port', tcp_port)

    @property
    def tcp_ports(self):
        return self.sports.keys()

    def write_to_serial(self, port, data):
        if port in self.sports:
            self.sports[port].write(data)

    def add_receiver(self, receiver):
        self.receivers.add(receiver)

    def remove_receiver(self, receiver):
        self.receivers.discard(receiver)


class SerialProtocol(asyncio.Protocol):
    worker = Worker()

    def connection_made(self, transport):
        self.transport = transport
        self.sock_port = self.transport.get_extra_info('sockname')[1]
        self.peer_host, self.peer_port = transport.get_extra_info('peername')
        self.receiver = Receiver(self.sock_port, self.peer_host, self.peer_port, self.transport)
        self.worker.add_receiver(self.receiver)
        print(f'Connection from {self.peer_host}:{self.peer_port}')

    def data_received(self, data):
        self.worker.write_to_serial(self.sock_port, data)

    def connection_lost(self, exc):
        print(f'Connection lost from {self.peer_host}:{self.peer_port}: {exc}')
        self.worker.remove_receiver(self.receiver)

    @staticmethod
    def tcp_ports():
        return SerialProtocol.worker.tcp_ports


def main():
    servers = []
    loop = asyncio.get_event_loop()
    for port in SerialProtocol.tcp_ports():
        servers.append(loop.run_until_complete(loop.create_server(SerialProtocol, '0.0.0.0', port)))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        print('Stopping and cleaning up...')
    finally:
        for server in servers:
            server.close()
            loop.run_until_complete(server.wait_closed())
        loop.close()


if __name__ == '__main__':
    main()
